/* Make camera move, reading keyboard.

TIPS:
- this is one possible object component.
- this class derive from CNodeComponent, see nodecomp.h for more info.
*/



#include <string.h>

#include "../defs.h"

#include "../mem.h"

#include "../array.h"

#include "../string.h"

#include "../node.h"

#include "../list.h"

#include "../polygon.h"

#include "../nodepol.h"

#include "../nodeint.h"

#include "../map.h"

#include "../math.h"

#include "../mesh.h"

#include "../sector.h"

#include "../nodesec.h"

#include "../refsys.h"

#include "../camsys.h"

#include "../nodesec.h"

#include "../lighttab.h"

#include "../eventtab.h"

#include "../scenery.h"

#include "../object.h"

#include "../nodeobj.h"

#include "nodecomp.h"

#include "cocamsys.h"

#include "comesh.h"

#include "comove.h"

#include "cokbcont.h"



/* constructor - set the type variable from CNodeComponent
**********************************************************/
CComponentKbControl::CComponentKbControl ()
{
   m_dwMyTypeID = TYPE_KBCONTROL;
}



/* Run the camera.
pData1 - not used
pData2 - not used
pData3 - not used
Return value 0 - ok / 1 - user press exit command
**********************************************************/
PTRVAL CComponentKbControl::mActivate ( PTRVAL pData1, PTRVAL pData2, PTRVAL pData3 )
{
   pData1 = pData1;  // to avoid GCC warnings
   pData2 = pData2;  // to avoid GCC warnings
   pData3 = pData3;  // to avoid GCC warnings

   // get the reference system
   CComponentCameraSystem
      *poComp = (CComponentCameraSystem*) m_poMyObj->mRetComponent ( SLOT_REFSYS );
   if ( poComp == NULL )
      ABORT ( MSG_ERROR_NO_TYPE, __FILE__, __LINE__ );
   CCameraSystem
      *poRefSys = &poComp->m_oCamSys;


   // define rotation angles (between 0 and 255)
   BYTE
      ayRot [ 3 ] = { 0, 0, 0 };     // axis x, y, z
   
   // acceleration direction
   tagfXYZ
      stAccel;
   stAccel.fX = 0;
   stAccel.fY = 0;
   stAccel.fZ = 0;


   // keyboard input

   // verify keyboard for Linux SVGALib target
   #ifdef TRG_LINUXSVGA   

   switch ( vga_getkey () ) {
      case 27:   // ESC
         // verify if need exit (!=91) or is a special key (==91)
         if ( vga_getkey () != 91 ) {
            m_poMyObj->m_poScen->m_bEnd = TRUE;
            return 1;
         }
         else
            switch ( vga_getkey () ) {
               case 'A':      // key UP
                  stAccel.fX += poRefSys->m_stJ.fX * poRefSys->m_fSpeed;
                  stAccel.fY += poRefSys->m_stJ.fY * poRefSys->m_fSpeed;
                  stAccel.fZ += poRefSys->m_stJ.fZ * poRefSys->m_fSpeed;
                  break;
               case 'B':      // key DOWN
                  stAccel.fX -= poRefSys->m_stJ.fX * poRefSys->m_fSpeed;
                  stAccel.fY -= poRefSys->m_stJ.fY * poRefSys->m_fSpeed;
                  stAccel.fZ -= poRefSys->m_stJ.fZ * poRefSys->m_fSpeed;
                  break;
               case 'C':      // key RIGHT
                  stAccel.fX += poRefSys->m_stI.fX * poRefSys->m_fSpeed;
                  stAccel.fY += poRefSys->m_stI.fY * poRefSys->m_fSpeed;
                  stAccel.fZ += poRefSys->m_stI.fZ * poRefSys->m_fSpeed;
                  break;
               case 'D':      // key LEFT
                  stAccel.fX -= poRefSys->m_stI.fX * poRefSys->m_fSpeed;
                  stAccel.fY -= poRefSys->m_stI.fY * poRefSys->m_fSpeed;
                  stAccel.fZ -= poRefSys->m_stI.fZ * poRefSys->m_fSpeed;
                  break;
            }
            break;

      // forward
      case 'A':
      case 'a':
         stAccel.fX += poRefSys->m_stK.fX * poRefSys->m_fSpeed;
         stAccel.fY += poRefSys->m_stK.fY * poRefSys->m_fSpeed;
         stAccel.fZ += poRefSys->m_stK.fZ * poRefSys->m_fSpeed;
         break;

      // backward
      case 'Z':
      case 'z':
         stAccel.fX -= poRefSys->m_stK.fX * poRefSys->m_fSpeed;
         stAccel.fY -= poRefSys->m_stK.fY * poRefSys->m_fSpeed;
         stAccel.fZ -= poRefSys->m_stK.fZ * poRefSys->m_fSpeed;
         break;

      // ROTATE
      case '8':       // key UP
         ayRot [ 0 ] -= 1;
         break;
      case '2':       // key DOWN
         ayRot [ 0 ] += 1;
         break;
      case '6':
         ayRot [ 1 ] -= 1;
         break;
      case '4':
         ayRot [ 1 ] += 1;
         break;
      case '7':
         ayRot [ 2 ] += 1;
         break;
      case '9':
         ayRot [ 2 ] -= 1;
         break;
   }
   #endif // END LINSVGA CODE


   // verify keyboard for DOS DJGPP or Linux X Window
   #if defined ( TRG_DOSDJGPP ) || defined ( TRG_LINUXXWIN )
   
   // test exit
   if ( m_oKb.mTestKeyPress ( K_ESC ) ) {
      m_poMyObj->m_poScen->m_bEnd = TRUE;
      return 1;
   }
   // test ALT
   if ( m_oKb.mTestKeyPress ( K_ALT ) || m_oKb.mTestKeyPress ( K_ALT | SPECIAL ) ) {
      // move
      if ( m_oKb.mTestKeyPress ( K_UP ) || m_oKb.mTestKeyPress ( K_UP | SPECIAL ) ) {
         stAccel.fX += poRefSys->m_stJ.fX * poRefSys->m_fSpeed;
         stAccel.fY += poRefSys->m_stJ.fY * poRefSys->m_fSpeed;
         stAccel.fZ += poRefSys->m_stJ.fZ * poRefSys->m_fSpeed;
      }
      if ( m_oKb.mTestKeyPress ( K_DOWN ) || m_oKb.mTestKeyPress ( K_DOWN | SPECIAL ) ) {
         stAccel.fX -= poRefSys->m_stJ.fX * poRefSys->m_fSpeed;
         stAccel.fY -= poRefSys->m_stJ.fY * poRefSys->m_fSpeed;
         stAccel.fZ -= poRefSys->m_stJ.fZ * poRefSys->m_fSpeed;
      }
      if ( m_oKb.mTestKeyPress ( K_RIGHT ) || m_oKb.mTestKeyPress ( K_RIGHT | SPECIAL ) ) {
         stAccel.fX += poRefSys->m_stI.fX * poRefSys->m_fSpeed;
         stAccel.fY += poRefSys->m_stI.fY * poRefSys->m_fSpeed;
         stAccel.fZ += poRefSys->m_stI.fZ * poRefSys->m_fSpeed;
      }
      if ( m_oKb.mTestKeyPress ( K_LEFT ) || m_oKb.mTestKeyPress ( K_LEFT | SPECIAL ) ) {
         stAccel.fX -= poRefSys->m_stI.fX * poRefSys->m_fSpeed;
         stAccel.fY -= poRefSys->m_stI.fY * poRefSys->m_fSpeed;
         stAccel.fZ -= poRefSys->m_stI.fZ * poRefSys->m_fSpeed;
      }
   }
   else {
      // rotations
      if ( m_oKb.mTestKeyPress ( K_UP ) || m_oKb.mTestKeyPress ( K_UP | SPECIAL ) )
         ayRot [ 0 ] -= 1;
      if ( m_oKb.mTestKeyPress ( K_DOWN ) || m_oKb.mTestKeyPress ( K_DOWN | SPECIAL ) )
         ayRot [ 0 ] += 1;
      if ( m_oKb.mTestKeyPress ( K_RIGHT ) || m_oKb.mTestKeyPress ( K_RIGHT | SPECIAL ) )
         ayRot [ 1 ] -= 1;
      if ( m_oKb.mTestKeyPress ( K_LEFT ) || m_oKb.mTestKeyPress ( K_LEFT | SPECIAL ) )
         ayRot [ 1 ] += 1;
   }
   if ( m_oKb.mTestKeyPress ( K_HOME ) )
      ayRot [ 2 ] += 1;
   if ( m_oKb.mTestKeyPress ( K_PG_UP ) )
      ayRot [ 2 ] -= 1;
   if ( m_oKb.mTestKeyPress ( K_A ) ) {
      stAccel.fX += poRefSys->m_stK.fX * poRefSys->m_fSpeed;
      stAccel.fY += poRefSys->m_stK.fY * poRefSys->m_fSpeed;
      stAccel.fZ += poRefSys->m_stK.fZ * poRefSys->m_fSpeed;
   }
   if ( m_oKb.mTestKeyPress ( K_Z ) ) {
      stAccel.fX -= poRefSys->m_stK.fX * poRefSys->m_fSpeed;
      stAccel.fY -= poRefSys->m_stK.fY * poRefSys->m_fSpeed;
      stAccel.fZ -= poRefSys->m_stK.fZ * poRefSys->m_fSpeed;
   }
   
   #endif // END   DJGPP and X Window    CODE


   // update move component
   // set direction acceleration
   m_poMyObj->mActivateComponent ( SLOT_MOVE, 3, (PTRVAL) &stAccel, 0 );
   // set rotation
   m_poMyObj->mActivateComponent ( SLOT_MOVE, 7, (PTRVAL) ayRot, 0 );
   
   // try make move/rotate
   m_poMyObj->mActivateComponent ( SLOT_MOVE, 2, 0, 0 );
   

   return 0;
}
