/* Flat poly with only 4 vertexes. */



#include <string.h>
#include <math.h>

#include "../defs.h"

#include "../mem.h"

#include "../string.h"

#include "../texture.h"

#include "../lighttab.h"

#include "../asm/draw.inc"

#include "draw.h"



// compute increase rate for x
#define INCREASEX( XI, YI, XF, YF ) ( (float) ( ( XF ) - ( XI ) ) / ( ( YF ) - ( YI ) ) )



/* Draw a flat polygon with only 4 vertex in frame buffer.
Medium optimized. The vertexes must be in clockwise or
anticlockwise. yCol is the polygon fill color.
This function use in implementation fix point 16.16.
Perform clip.

                        (x4, y4)
                        *                -
                       / \                 loop
                      /   \                    |
            (x1, y1) *     \                   |
                      \     * (x3, y3)         | compute x positions
                       \   /                   | between edges
                        \ /                    V
                         *               -
                         (x2, y2)

Line equation:
Y = aX + b, where
            a is a increase rate
            b is the offset of the line in y axis
Equation used in intersections (how x change for each scanline + x vertex)
X = ( deltaY * increaseX ) + Xvertex, where
            increaseX is 1 / a, how x change per scanline
            deltaY is the distance between a Y vertex and Y loop
            Xvertex is the X in vertex used in computation of deltaY
**********************************************************/
void CDraw::mFlatPolygon4 ( INT nX1, INT nY1, INT nX2, INT nY2,
                            INT nX3, INT nY3, INT nX4, INT nY4, BYTE yCol )
{
   // clip y edges
   // define less (top) and greater (bottom) y values
   INT
      nYSup = nY1,
      nYInf = nY1;
   // test y2
   if ( nYSup > nY2 )
      nYSup = nY2;
   else
      if ( nYInf < nY2 )
         nYInf = nY2;
   // test y3
   if ( nYSup > nY3 )
      nYSup = nY3;
   else
      if ( nYInf < nY3 )
         nYInf = nY3;
   // test y4
   if ( nYSup > nY4 )
      nYSup = nY4;
   else
      if ( nYInf < nY4 )
         nYInf = nY4;
   // test if points are outside frame buffer limits, if yes change values
   if ( nYSup < 0 )
      nYSup = 0;
   if ( nYInf > m_nMaxY )
      nYInf = m_nMaxY;

   // define if be possible compute x increase rates, if yes do
   INT
      nIncX12 = 0,
      nIncX14 = 0,
      nIncX23 = 0,
      nIncX34 = 0;
   BOOL
      bIncValid12 = TRUE,
      bIncValid14 = TRUE,
      bIncValid23 = TRUE,
      bIncValid34 = TRUE;
   // be computed valores only for points that are not in same lines
   if ( nY1 != nY2 )
      nIncX12 = (INT) ( INCREASEX ( nX1, nY1, nX2, nY2 ) * 65536 );
   else
      bIncValid12 = FALSE;
   if ( nY1 != nY4 )
      nIncX14 = (INT) ( INCREASEX ( nX1, nY1, nX4, nY4 ) * 65536 );
   else
      bIncValid14 = FALSE;
   if ( nY2 != nY3 )
      nIncX23 = (INT) ( INCREASEX ( nX2, nY2, nX3, nY3 ) * 65536 );
   else
      bIncValid23 = FALSE;
   if ( nY3 != nY4 )
      nIncX34 = (INT) ( INCREASEX ( nX3, nY3, nX4, nY4 ) * 65536 );
   else
      bIncValid34 = FALSE;

   // define address of the beging for the first line in frame buffer
   UINT
      pLineAddress = (PTRVAL) m_poFrBuff->m_payMem + m_nResX * nYSup;

   // define color fill
   UINT
      uCol = yCol | ( yCol << 8 ) | ( yCol << 16 ) | ( yCol << 24 );

   // goto polygon start to end in Y axis
   INT
      nXIni,    // start x in scanline
      nXEnd,    // end x in scanline
      nX;       // x value of intersection by 2 lines
   for ( INT ncY = nYSup; ncY <= nYInf; ncY ++ ) {

      // initial x values in scanline
      nXIni = 32767;
      nXEnd = - 32768;

      // compute x between vertex 1 and 4
      if ( bIncValid14 &&
           ( ncY >= nY1 && ncY <= nY4 ) ||
           ( ncY <= nY1 && ncY >= nY4 ) ) {

         nX = (INT) ( ( ncY - nY4 ) * nIncX14 >> 16 ) + nX4;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // compute x between vertex 1 and 2
      if ( bIncValid12 &&
           ( ncY >= nY1 && ncY <= nY2 ) ||
           ( ncY <= nY1 && ncY >= nY2 ) ) {

         nX = (INT) ( ( ncY - nY2 ) * nIncX12 >> 16 ) + nX2;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // compute x between vertex 2 and 3
      if ( bIncValid23 &&
           ( ncY >= nY2 && ncY <= nY3 ) ||
           ( ncY <= nY2 && ncY >= nY3 ) ){

         nX = (INT) ( ( ncY - nY3 ) * nIncX23 >> 16 ) + nX3;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // compute x between vertex 3 and 4
      if ( bIncValid34 &&
           ( ncY >= nY3 && ncY <= nY4 ) ||
           ( ncY <= nY3 && ncY >= nY4 ) ) {

         nX = (INT) ( ( ncY - nY4 ) * nIncX34 >> 16 ) + nX4;

         // verify if start x must be changed
         if ( nX < nXIni )
            nXIni = nX;
         // verify if end x must be changed
         if ( nX > nXEnd )
            nXEnd = nX;
      }

      // clip x values
      // verify start
      if ( nXIni < 0 )
         nXIni = 0;
      else
         // start can be in right (line outside frame buffer)
         if ( nXIni > m_nMaxX ) {
            // compute next scanline address
            pLineAddress += m_nResX;
            // do not draw line
            continue;
         }
      // verify end
      if ( nXEnd > m_nMaxX )
         nXEnd = m_nMaxX;
      else
         // end can be in left (line outside frame buffer)
         if ( nXEnd < 0 ) {
            // compute next scanline address
            pLineAddress += m_nResX;
            // do not draw line
            continue;
         }

      // draw horizontal line

      // if compile with asm code
      #ifdef COMP_ASM
      ASM_HLINE_FLAT ( pLineAddress + nXIni, nXEnd - nXIni + 1, uCol );
      #endif

      // compile C only (no asm code)
      #ifndef COMP_ASM
      memset ( (void*) ( pLineAddress + nXIni ), uCol, nXEnd - nXIni + 1 );
      #endif

      // compute next scanline address
      pLineAddress += m_nResX;
   }
}

